//This file is part of the Attempto Java Packages.
//Copyright 2008-2009, Attempto Group, University of Zurich (see http://attempto.ifi.uzh.ch).

//The Attempto Java Packages is free software: you can redistribute it and/or modify it under the
//terms of the GNU Lesser General Public License as published by the Free Software Foundation,
//either version 3 of the License, or (at your option) any later version.

//The Attempto Java Packages is distributed in the hope that it will be useful, but WITHOUT ANY
//WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
//PURPOSE. See the GNU Lesser General Public License for more details.

//You should have received a copy of the GNU Lesser General Public License along with the Attempto
//Java Packages. If not, see http://www.gnu.org/licenses/.

package ch.uzh.ifi.attempto.aceeditor;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.Properties;

import nextapp.echo2.app.ApplicationInstance;
import nextapp.echo2.app.Column;
import nextapp.echo2.app.Component;
import nextapp.echo2.app.Extent;
import nextapp.echo2.app.Insets;
import nextapp.echo2.app.SplitPane;
import nextapp.echo2.app.Window;
import nextapp.echo2.app.event.ActionEvent;
import nextapp.echo2.app.event.ActionListener;
import nextapp.echo2.app.filetransfer.Download;
import nextapp.echo2.app.filetransfer.DownloadProvider;
import nextapp.echo2.webcontainer.command.BrowserRedirectCommand;
import ch.uzh.ifi.attempto.aceeditor.grammar.ACEGrammar;
import ch.uzh.ifi.attempto.echocomp.MessageWindow;
import ch.uzh.ifi.attempto.echocomp.TextAreaWindow;
import ch.uzh.ifi.attempto.echocomp.UploadWindow;
import ch.uzh.ifi.attempto.preditor.PreditorWindow;
import ch.uzh.ifi.attempto.preditor.text.ContextChecker;
import ch.uzh.ifi.attempto.preditor.text.EnglishContextChecker;
import ch.uzh.ifi.attempto.preditor.text.TextContainer;
import ch.uzh.ifi.attempto.preditor.text.TextElement;
import echopointng.KeyStrokeListener;

/**
 * This is the ACE Editor web application.
 * 
 * @author Tobias Kuhn
 */
public class ACEEditor extends Window implements ActionListener {

	private static final long serialVersionUID = -684743065195237612L;

	private static final ContextChecker contextChecker = new EnglishContextChecker(true, true);

	private Column textColumn = new Column();
	private Column mainColumn = new Column();
	private MenuBar menuBar;
	private TextEntry selectedEntry;
	private TextEntry finalEntry = new TextEntry(null, this);
	private boolean edit;
	private TextEntry clipboard;
	private KeyStrokeListener keyStrokeListener = new KeyStrokeListener();
	private int downloadCount = 0;
	private LexiconHandler lexiconHandler;
	private boolean parseWithClexEnabled;

	private static Properties properties;

	/**
	 * Creates a new ACE Editor application.
	 * 
	 * @param lexiconFile The file name of the lexicon to be used.
	 * @param parseWithClexEnabled true if the compiled lexicon (clex) of the APE executable should be used for
	 *    parsing the ACE sentences. This should be done only if the compiled lexicon is the same as the
	 *    lexicon defined by the first argument. Parsing with the compiled lexicon is faster.
	 */
	public ACEEditor(String lexiconFile, boolean parseWithClexEnabled) {
		setTitle("ACE Editor");
		lexiconHandler = new LexiconHandler(lexiconFile);
		this.parseWithClexEnabled = parseWithClexEnabled;

		SplitPane splitPane = new SplitPane(SplitPane.ORIENTATION_VERTICAL);
		splitPane.setSeparatorPosition(new Extent(25));

		menuBar = new MenuBar(this);
		menuBar.setSelected("Default Expanded", true);
		menuBar.setSelected("Default Paraphrase", true);
		menuBar.setSelected("Default Syntax Boxes", true);
		menuBar.setSelected("Default Pretty-Printed DRS", true);
		menuBar.setEnabled("Paste", false);
		splitPane.add(menuBar);

		textColumn.setInsets(new Insets(0, 5));
		textColumn.add(finalEntry);

		mainColumn.add(textColumn);

		// Up and down keys for moving the selection:
		keyStrokeListener.addKeyCombination(KeyStrokeListener.VK_UP, "Up Pressed");
		keyStrokeListener.addKeyCombination(KeyStrokeListener.VK_DOWN, "Down Pressed");

		// Space key for expand/collapse or add:
		keyStrokeListener.addKeyCombination(KeyStrokeListener.VK_SPACE, "Space Pressed");

		// Backspace key for delete:
		keyStrokeListener.addKeyCombination(KeyStrokeListener.VK_BACK_SPACE, "Backspace Pressed");

		// Function key + A for add:
		keyStrokeListener.addKeyCombination(KeyStrokeListener.VK_A | KeyStrokeListener.CONTROL_MASK, "Func-A Pressed");
		keyStrokeListener.addKeyCombination(KeyStrokeListener.VK_A | KeyStrokeListener.META_MASK, "Func-A Pressed");
		keyStrokeListener.addKeyCombination(KeyStrokeListener.VK_A | KeyStrokeListener.ALT_MASK, "Func-A Pressed");

		// Function key + M for modify:
		keyStrokeListener.addKeyCombination(KeyStrokeListener.VK_M | KeyStrokeListener.CONTROL_MASK, "Func-M Pressed");
		keyStrokeListener.addKeyCombination(KeyStrokeListener.VK_M | KeyStrokeListener.META_MASK, "Func-M Pressed");
		keyStrokeListener.addKeyCombination(KeyStrokeListener.VK_M | KeyStrokeListener.ALT_MASK, "Func-M Pressed");

		// Function key + X for cut:
		keyStrokeListener.addKeyCombination(KeyStrokeListener.VK_X | KeyStrokeListener.CONTROL_MASK, "Func-X Pressed");
		keyStrokeListener.addKeyCombination(KeyStrokeListener.VK_X | KeyStrokeListener.META_MASK, "Func-X Pressed");
		keyStrokeListener.addKeyCombination(KeyStrokeListener.VK_X | KeyStrokeListener.ALT_MASK, "Func-X Pressed");

		// Function key + C for copy:
		keyStrokeListener.addKeyCombination(KeyStrokeListener.VK_C | KeyStrokeListener.CONTROL_MASK, "Func-C Pressed");
		keyStrokeListener.addKeyCombination(KeyStrokeListener.VK_C | KeyStrokeListener.META_MASK, "Func-C Pressed");
		keyStrokeListener.addKeyCombination(KeyStrokeListener.VK_C | KeyStrokeListener.ALT_MASK, "Func-C Pressed");

		// Function key + V for paste:
		keyStrokeListener.addKeyCombination(KeyStrokeListener.VK_V | KeyStrokeListener.CONTROL_MASK, "Func-V Pressed");
		keyStrokeListener.addKeyCombination(KeyStrokeListener.VK_V | KeyStrokeListener.META_MASK, "Func-V Pressed");
		keyStrokeListener.addKeyCombination(KeyStrokeListener.VK_V | KeyStrokeListener.ALT_MASK, "Func-V Pressed");

		// Function key + O for open:
		keyStrokeListener.addKeyCombination(KeyStrokeListener.VK_O | KeyStrokeListener.CONTROL_MASK, "Func-O Pressed");
		keyStrokeListener.addKeyCombination(KeyStrokeListener.VK_O | KeyStrokeListener.META_MASK, "Func-O Pressed");
		keyStrokeListener.addKeyCombination(KeyStrokeListener.VK_O | KeyStrokeListener.ALT_MASK, "Func-O Pressed");

		// Function key + S for save:
		keyStrokeListener.addKeyCombination(KeyStrokeListener.VK_S | KeyStrokeListener.CONTROL_MASK, "Func-S Pressed");
		keyStrokeListener.addKeyCombination(KeyStrokeListener.VK_S | KeyStrokeListener.META_MASK, "Func-S Pressed");
		keyStrokeListener.addKeyCombination(KeyStrokeListener.VK_S | KeyStrokeListener.ALT_MASK, "Func-S Pressed");

		keyStrokeListener.addActionListener(this);
		mainColumn.add(keyStrokeListener);

		splitPane.add(mainColumn);
		getContent().add(splitPane);

		select(finalEntry);
	}

	/**
	 * Returns whether parsing with the compiled lexicon of the APE executable is enabled.
	 * 
	 * @return true if parsing with the compiled lexicon is enabled.
	 */
	public boolean isParseWithClexEnabled() {
		return parseWithClexEnabled;
	}

	LexiconHandler getLexiconHandler() {
		return lexiconHandler;
	}

	void select(TextEntry entry) {
		if (selectedEntry != null) {
			selectedEntry.setSelected(false);
		}
		entry.setSelected(true);
		selectedEntry = entry;

		if (selectedEntry == finalEntry) {
			menuBar.setEnabled("Delete", false);
			menuBar.setEnabled("Cut", false);
		} else {
			menuBar.setEnabled("Delete", true);
			menuBar.setEnabled("Cut", true);
		}
		if (selectedEntry.isEmpty()) {
			menuBar.setEnabled("Modify...", false);
		} else {
			menuBar.setEnabled("Modify...", true);
		}
		if (selectedEntry.isEmpty() || selectedEntry.hasError() || selectedEntry.isComment()) {
			menuBar.setEnabled("Expanded", false);
			menuBar.setSelected("Expanded", false);
			for (String s : ResultItem.TYPES) {
				menuBar.setEnabled("Show " + s, false);
				menuBar.setSelected("Show " + s, false);
			}
		} else {
			menuBar.setEnabled("Expanded", true);
			menuBar.setSelected("Expanded", selectedEntry.isExpanded());
			for (String s : ResultItem.TYPES) {
				menuBar.setEnabled("Show " + s, true);
				menuBar.setSelected("Show " + s, selectedEntry.isResultItemVisible(s));
			}
		}
		menuBar.update();
	}

	public void actionPerformed(ActionEvent e) {
		String command = e.getActionCommand();
		Object source = e.getSource();

		if (command.equals("About")) {
			String v = getInfo("aceeditor-version") + " (" + getInfo("aceeditor-release-stage") + ")";
			String d = getInfo("aceeditor-build-date");
			String a = getInfo("aceeditor-developer");
			getContent().add(new MessageWindow("ACE Editor", "ACE Editor " + v + ", " + d + ", " + a, "OK"));
		} else if (command.equals("Attempto Website")) {
			ApplicationInstance.getActive().enqueueCommand(new BrowserRedirectCommand("http://attempto.ifi.uzh.ch"));
		} else if (command.equals("Open...")) {
			openFile();
		} else if (command.equals("Save As...")) {
			saveFile();
		} else if (command.equals("Add...")) {
			showEditor(false);
		} else if (command.equals("Add Comment...")) {
			showCommentEditor(false);
		} else if (command.equals("Add Separator")) {
			TextEntry newEntry = new TextEntry(null, this);
			textColumn.add(newEntry, textColumn.indexOf(selectedEntry));
			select(newEntry);
		} else if (command.equals("Modify...")) {
			if (selectedEntry.isComment()) {
				showCommentEditor(true);
			} else {
				showEditor(true);
			}
		} else if (command.equals("Delete")) {
			deleteSelectedEntry();
		} else if (command.equals("Cut")) {
			cutSelectedEntry();
		} else if (command.equals("Copy")) {
			copySelectedEntry();
		} else if (command.equals("Paste")) {
			pasteFromClipboard();
		} else if (command.equals("Expanded")) {
			selectedEntry.setExpanded(menuBar.isSelected("Expanded"));
		} else if (command.equals("Expand All")) {
			for (Component c : textColumn.getComponents()) {
				((TextEntry) c).setExpanded(true);
			}
		} else if (command.equals("Collapse All")) {
			for (Component c : textColumn.getComponents()) {
				((TextEntry) c).setExpanded(false);
			}
		} else if (command.startsWith("Show ")) {
			selectedEntry.setResultItemVisible(command.substring(5), menuBar.isSelected(command));
			selectedEntry.setExpanded(true);
		} else if (source instanceof PreditorWindow && command.equals("Cancel")) {
			PreditorWindow preditor = (PreditorWindow) source;
			preditor.setVisible(false);
			preditor.dispose();
			refreshKeyStrokeListener();
		} else if (source instanceof PreditorWindow && command.equals("OK")) {
			PreditorWindow preditor = (PreditorWindow) source;
			preditor.setContextChecker(contextChecker);
			TextContainer textContainer = preditor.getTextContainer();
			if (textContainer.getTextElementsCount() == 0) {
				preditor.dispose();
				preditor.setVisible(false);
				refreshKeyStrokeListener();
			} else {
				ArrayList<TextElement> finalElements = preditor.getPossibleNextTokens(".", "?");
				if (!finalElements.isEmpty()) textContainer.addElement(finalElements.get(0));
				ArrayList<TextElement> l = textContainer.getTextElements();
				if (l.isEmpty() || l.get(l.size() - 1).getText().matches("[.?]")) {
					if (edit) {
						selectedEntry.setAceText(textContainer.getText());
						select(selectedEntry);
					} else {
						TextEntry newEntry = new TextEntry(textContainer.getText(), this, menuBar.isSelected("Default Expanded"));
						for (String s : ResultItem.TYPES) {
							newEntry.setResultItemVisible(s, menuBar.isSelected("Default " + s));
						}
						textColumn.add(newEntry, textColumn.indexOf(selectedEntry));
						select(newEntry);
					}
					preditor.dispose();
					preditor.setVisible(false);
					refreshKeyStrokeListener();
				} else {
					getContent().add(new MessageWindow("Error", "There are unfinished sentences.", "OK"));
				}
			}
		} else if (source instanceof TextAreaWindow && command.equals("OK")) {
			TextAreaWindow cew = (TextAreaWindow) source;
			if (edit) {
				selectedEntry.setAceText("# " + cew.getText());
				select(selectedEntry);
			} else {
				TextEntry newEntry = new TextEntry("# " + cew.getText(), this, false);
				textColumn.add(newEntry, textColumn.indexOf(selectedEntry));
				select(newEntry);
			}
		} else if (command.equals("Upload")) {
			String fileContent = ((UploadWindow) source).getFileContent();
			if (fileContent != null) {
				textColumn.removeAll();
				String[] l = fileContent.replaceAll("\\s*(#[^\\n]*\\n)", "\n\n$1\n").split("\\n[ \\t\\x0B\\f\\r]*\\n");
				for (String line : l) {
					TextEntry newEntry = new TextEntry(line, this, false);
					textColumn.add(newEntry);
					for (String s : ResultItem.TYPES) {
						newEntry.setResultItemVisible(s, menuBar.isSelected("Default " + s));
					}
				}
				textColumn.add(finalEntry);
				select((TextEntry) textColumn.getComponent(0));
			}
		} else if (command.equals("Up Pressed")) {
			int i = textColumn.indexOf(selectedEntry);
			if (i > 0) {
				select((TextEntry) textColumn.getComponent(i-1));
			}
		} else if (command.equals("Down Pressed")) {
			int i = textColumn.indexOf(selectedEntry);
			if (i < textColumn.getComponentCount()-1) {
				select((TextEntry) textColumn.getComponent(i+1));
			}
		} else if (command.equals("Space Pressed")) {
			if (selectedEntry.isEmpty()) {
				showEditor(false);
			} else {
				if (selectedEntry.isExpanded()) {
					selectedEntry.setExpanded(false);
				} else {
					selectedEntry.setExpanded(true);
				}
			}
		} else if (command.equals("Backspace Pressed")) {
			deleteSelectedEntry();
		} else if (command.equals("Func-A Pressed")) {
			showEditor(false);
		} else if (command.equals("Func-M Pressed")) {
			if (selectedEntry.isComment()) {
				showCommentEditor(true);
			} else {
				showEditor(true);
			}
		} else if (command.equals("Func-X Pressed")) {
			cutSelectedEntry();
		} else if (command.equals("Func-C Pressed")) {
			copySelectedEntry();
		} else if (command.equals("Func-V Pressed")) {
			pasteFromClipboard();
		} else if (command.equals("Func-O Pressed")) {
			openFile();
		} else if (command.equals("Func-S Pressed")) {
			saveFile();
		}
	}

	private void refreshKeyStrokeListener() {
		// The different keystroke listeners somehow interfere with each other so that this
		// work-around is needed:
		mainColumn.remove(keyStrokeListener);
		mainColumn.add(keyStrokeListener);
	}

	private void showEditor(boolean edit) {
		if (edit && selectedEntry.isEmpty()) return;

		PreditorWindow preditor = new PreditorWindow("ACE Text Editor", new ACEGrammar(), new ACEEditorMenuCreator(lexiconHandler));
		preditor.setContextChecker(contextChecker);
		preditor.addActionListener(this);
		this.edit = edit;
		if (edit) {
			preditor.addText(selectedEntry.getAceText() + " ");
		}
		getContent().add(preditor);
	}

	private void showCommentEditor(boolean edit) {
		this.edit = edit;
		if (edit) {
			getContent().add(new TextAreaWindow("Comment Editor", selectedEntry.getAceText().substring(2), this));
		} else {
			getContent().add(new TextAreaWindow("Comment Editor", "", this));
		}
	}

	private void deleteSelectedEntry() {
		if (selectedEntry != finalEntry) {
			int i = textColumn.indexOf(selectedEntry);
			TextEntry nextEntry = (TextEntry) textColumn.getComponent(i+1);
			textColumn.remove(selectedEntry);
			select(nextEntry);
		}
	}

	private void copySelectedEntry() {
		clipboard = selectedEntry.copy();
		menuBar.setEnabled("Paste", true);
		menuBar.update();
	}

	private void cutSelectedEntry() {
		if (selectedEntry != finalEntry) {
			copySelectedEntry();
			deleteSelectedEntry();
		}
	}

	private void pasteFromClipboard() {
		if (clipboard != null) {
			TextEntry newEntry = clipboard.copy();
			textColumn.add(newEntry, textColumn.indexOf(selectedEntry));
			select(newEntry);
		}
	}

	private void openFile() {
		UploadWindow uw = new UploadWindow(
				"Open File",
				"Warning: This will delete the current content.\nChoose a file to open:",
				null,
				this
		);
		uw.setMaxFileSize(10000);
		getContent().add(uw);
	}

	private void saveFile() {
		final String f = getFullText();
		downloadCount++;
		DownloadProvider provider = new DownloadProvider() {

			public String getContentType() {
				return "text/plain";
			}

			public String getFileName() {
				return "text" + downloadCount + ".ace.txt";
			}

			public int getSize() {
				return f.length();
			}

			public void writeFile(OutputStream out) throws IOException {
				out.write(f.getBytes());
				out.close();
			}

		};
		getApplicationInstance().enqueueCommand(new Download(provider, true));
	}

	void entryChanged(TextEntry entry) {
		if (entry == selectedEntry) {
			menuBar.setSelected("Expanded", selectedEntry.isExpanded());
			menuBar.update();
		}
	}

	/**
	 * Returns the full text of the current content of this ACE Editor instance.
	 * 
	 * @return The full text.
	 */
	public String getFullText() {
		String text = "";
		for (Component c : textColumn.getComponents()) {
			String s = ((TextEntry) c).getAceText();
			if (c == finalEntry) break;
			if (s == null) s = "";
			text += s + "\n\n";
		}
		return text;
	}

	/**
	 * Returns information about ACE Editor, like the version number and the release date. This
	 * information is read from the file "aceeditor.properties".
	 * 
	 * @param key The key string.
	 * @return The value for the given key.
	 */
	public static String getInfo(String key) {
		if (properties == null) {
			String f = "ch/uzh/ifi/attempto/aceeditor/aceeditor.properties";
			InputStream in = Thread.currentThread().getContextClassLoader().getResourceAsStream(f);
			properties = new Properties();
			try {
				properties.load(in);
			} catch (Exception ex) {
				ex.printStackTrace();
			}
		}

		return properties.getProperty(key);
	}

}
