// This file is part of the Attempto Java Packages.
// Copyright 2008-2009, Attempto Group, University of Zurich (see http://attempto.ifi.uzh.ch).
//
// The Attempto Java Packages is free software: you can redistribute it and/or modify it under the
// terms of the GNU Lesser General Public License as published by the Free Software Foundation,
// either version 3 of the License, or (at your option) any later version.
//
// The Attempto Java Packages is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE. See the GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License along with the Attempto
// Java Packages. If not, see http://www.gnu.org/licenses/.

package ch.uzh.ifi.attempto.acewiki.gui.editor;

import nextapp.echo2.app.event.ActionEvent;
import nextapp.echo2.app.event.ActionListener;
import ch.uzh.ifi.attempto.acewiki.Wiki;
import ch.uzh.ifi.attempto.acewiki.core.ontology.Individual;
import ch.uzh.ifi.attempto.acewiki.core.ontology.OntologyElement;
import ch.uzh.ifi.attempto.acewiki.core.ontology.OntologyTextElement;
import ch.uzh.ifi.attempto.chartparser.Restriction;
import ch.uzh.ifi.attempto.chartparser.Terminal;
import ch.uzh.ifi.attempto.preditor.MenuCreator;
import ch.uzh.ifi.attempto.preditor.MenuEntry;
import ch.uzh.ifi.attempto.preditor.SpecialMenuItem;
import ch.uzh.ifi.attempto.preditor.text.TextElement;

/**
 * This is the menu creator class that generates the menu entries for the predictive editor
 * on the basis of the AceWiki grammar.
 * 
 * @author Tobias Kuhn
 */
class AceWikiMenuCreator extends MenuCreator implements ActionListener {
	
	private static final long serialVersionUID = -110533069194755735L;
	
	private Wiki wiki;
	private OntologyElement highlightedElement;
	private ActionListener actionListener;
	
	/**
	 * Creates a new AceWiki-specific menu creator object.
	 * 
	 * @param wiki The wiki instance.
	 * @param highlightedElement The ontology element that should be highlighted in the editor (because it is
	 *   the current element).
	 * @param actionListener The action-listener.
	 */
	public AceWikiMenuCreator(Wiki wiki, OntologyElement highlightedElement, ActionListener actionListener) {
		this.wiki = wiki;
		this.highlightedElement = highlightedElement;
		this.actionListener = actionListener;
	}
	
	public void initMenuCreation() {
		prepareMenuBlock("function word", true);
		prepareMenuBlock("noun", true);
		prepareMenuBlock("plural noun", true);
		prepareMenuBlock("proper name", true);
		prepareMenuBlock("of-construct", true);
		prepareMenuBlock("verb", true);
		prepareMenuBlock("passive verb", true);
		prepareMenuBlock("transitive adjective", true);
		prepareMenuBlock("reference", false);
		prepareMenuBlock("new variable", false);
	}

	public void addMenuItems(Restriction restriction) {
		String n = restriction.getCategory().getName();
		
		if (n.equals("PROPERNAME")) {
			addMenuItem("proper name", new SpecialMenuItem("new...", "new propername", this));
			addMenuEntries("proper name", "Proper Name", 0);
			// abbreviations have to be added separately:
			for (OntologyElement el : wiki.getOntologyElements()) {
				if (!(el instanceof Individual)) continue;
				Individual ind = (Individual) el;
				if (ind.getAbbreviation() == null) continue;
				MenuEntry menuEntry = new MenuEntry(OntologyTextElement.createTextElement(el, 2));
				menuEntry.setHighlighted(el == highlightedElement);
				addMenuItem("proper name", menuEntry);
			}
		} else if (n.equals("NOUN")) {
			addMenuItem("noun", new SpecialMenuItem("new...", "new noun", this));
			addMenuEntries("noun", "Noun", 0);
		} else if (n.equals("DEF_NOUN")) {
			for (OntologyElement el : wiki.getOntologyElements()) {
				if (!"Noun".equals(el.getType())) continue;
				if (el.getWord(0) == null) continue;
				Terminal cat = new Terminal("DEF_NOUN");
				cat.setFeature("noun", el.getWord(0));
				cat.setFeature("text", "the " + el.getWord(0));
				addMenuItem("reference", new MenuEntry(new TextElement("the " + el.getWord(0), cat)));
			}
		} else if (n.equals("NOUNPL")) {
			addMenuItem("plural noun", new SpecialMenuItem("new...", "new nounpl", this));
			addMenuEntries("plural noun", "Noun", 1);
		} else if (n.equals("NOUNOF")) {
			addMenuItem("of-construct", new SpecialMenuItem("new...", "new nounof", this));
			addMenuEntries("of-construct", "Of-Construct", 0);
		} else if (n.equals("VERB")) {
			addMenuItem("verb", new SpecialMenuItem("new...", "new verb", this));
			addMenuEntries("verb", "Verb", 0);
		} else if (n.equals("VERBINF")) {
			addMenuItem("verb", new SpecialMenuItem("new...", "new verbinf", this));
			addMenuEntries("verb", "Verb", 1);
		} else if (n.equals("PVERB")) {
			addMenuItem("passive verb", new SpecialMenuItem("new...", "new pverb", this));
			addMenuEntries("passive verb", "Verb", 2);
		} else if (n.equals("TRADJ")) {
			addMenuItem("transitive adjective", new SpecialMenuItem("new...", "new tradj", this));
			addMenuEntries("transitive adjective", "Transitive Adjective", 0);
		} else if (n.equals("VARIABLE")) {
			addVariableEntries("new variable", 5, "VARIABLE");
		} else if (n.equals("REFERENCE")) {
			addVariableEntries("reference", 5, "REFERENCE");
		} else if (n.equals("NUMBER")) {
			for (int i = 2; i <= 100; i++) {
				addMenuEntry("function word", new TextElement(i + "", "NUMBER"));
			}
		} else {
			addMenuEntry("function word", new TextElement(n));
		}
	}
	
	private void addMenuEntries(String menuBlockName, String type, int wordNumber) {
		for (OntologyElement el : wiki.getOntologyElements()) {
			if (!type.equals(el.getType())) continue;
			if (el.getWord(wordNumber) == null) continue;
			MenuEntry menuEntry = new MenuEntry(OntologyTextElement.createTextElement(el, wordNumber));
			menuEntry.setHighlighted(el == highlightedElement);
			addMenuItem(menuBlockName, menuEntry);
		}
	}
	
	public void actionPerformed(ActionEvent e) {
		if (e.getActionCommand().equals("new propername")) {
			wiki.showWindow(ProperNameForm.createCreatorWindow(wiki, actionListener));
		} else if (e.getActionCommand().equals("new noun")) {
			wiki.showWindow(NounForm.createCreatorWindow(0, wiki, actionListener));
		} else if (e.getActionCommand().equals("new nounpl")) {
			wiki.showWindow(NounForm.createCreatorWindow(1, wiki, actionListener));
		} else if (e.getActionCommand().equals("new nounof")) {
			wiki.showWindow(NounOfForm.createCreatorWindow(wiki, actionListener));
		} else if (e.getActionCommand().equals("new verb")) {
			wiki.showWindow(VerbForm.createCreatorWindow(0, wiki, actionListener));
		} else if (e.getActionCommand().equals("new verbinf")) {
			wiki.showWindow(VerbForm.createCreatorWindow(1, wiki, actionListener));
		} else if (e.getActionCommand().equals("new pverb")) {
			wiki.showWindow(VerbForm.createCreatorWindow(2, wiki, actionListener));
		} else if (e.getActionCommand().equals("new tradj")) {
			wiki.showWindow(TrAdjForm.createCreatorWindow(wiki, actionListener));
		}
	}

}
