// This file is part of the Attempto Java Packages.
// Copyright 2008-2009, Attempto Group, University of Zurich (see http://attempto.ifi.uzh.ch).
//
// The Attempto Java Packages is free software: you can redistribute it and/or modify it under the
// terms of the GNU Lesser General Public License as published by the Free Software Foundation,
// either version 3 of the License, or (at your option) any later version.
//
// The Attempto Java Packages is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE. See the GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License along with the Attempto
// Java Packages. If not, see http://www.gnu.org/licenses/.

package ch.uzh.ifi.attempto.chartparser;

import java.util.HashMap;

/**
 * This class represents a grammatical category.
 * 
 * @author Tobias Kuhn
 */
public abstract class Category {
	
	private final String name;
	private FeatureMap featureMap = new FeatureMap();
	
	/**
	 * Creates a new category.
	 * 
	 * @param name The name of the category.
	 */
	public Category(String name) {
		this.name = name;
	}
	
	/**
	 * Returns true if this category is a reference.
	 * 
	 * @return true if this is a reference.
	 */
	public boolean isReference() {
		if (name.equals(">>>")) return true;
		if (name.equals("<<<")) return true;
		if (name.equals("</<")) return true;
		return false;
	}
	
	/**
	 * Returns the name of the category.
	 * 
	 * @return The name of the category.
	 */
	public String getName() {
		return name;
	}
	
	/**
	 * Returns the map of features of this category.
	 * 
	 * @return The feature map.
	 */
	public FeatureMap getFeatureMap() {
		return featureMap;
	}
	
	/**
	 * Sets a feature.
	 * 
	 * @param featureName The feature name
	 * @param featureValue The string reference that points to the value of the feature.
	 */
	public void setFeature(String featureName, StringRef featureValue) {
		featureMap.setFeature(featureName, featureValue);
	}
	
	/**
	 * Sets a feature.
	 * 
	 * @param featureName The feature name
	 * @param featureValue The value of the feature.
	 */
	public void setFeature(String featureName, String featureValue) {
		featureMap.setFeature(featureName, new StringRef(featureValue));
	}
	
	/**
	 * Returns a feature value.
	 * 
	 * @param featureName The name of the feature.
	 * @return The value of the feature.
	 */
	public StringRef getFeature(String featureName) {
		return featureMap.getFeature(featureName);
	}
	
	/**
	 * Unifies this category with another category. Two categories can unify if and only if they have
	 * the same names and they have no features with conflicting values. If the unification fails, a
	 * UnificationFailedException is thrown. In this case, the two categories remain partly unified,
	 * i.e. no backtracking is done. Thus, this operation should be perfomed only if it is certain that
	 * the unification succeeds, or if the operation is performed on copies of objects that are not
	 * used anymore afterwards.
	 * 
	 * @param category The category to be unified with this category.
	 * @throws UnificationFailedException If unification fails.
	 */
	public void unify(Category category) throws UnificationFailedException {
		if (!name.equals(category.name)) {
			throw new UnificationFailedException();
		}
		featureMap.unify(category.featureMap);
	}
	
	/**
	 * Tries to unify this category with another category. If unification is not possible, an exception
	 * is thrown. In the case unification would be possible, the unification is not performed completely.
	 * In any case the two categories remain in an unconsistent state afterwards. Thus, this operation should
	 * be performed only on copies of objects that are not used anymore afterwards.
	 * 
	 * @param category The category to be unified with this category.
	 * @throws UnificationFailedException If unification fails.
	 */
	public void tryToUnify(Category category) throws UnificationFailedException {
		if (!name.equals(category.name)) {
			throw new UnificationFailedException();
		}
		featureMap.tryToUnify(category.featureMap);
	}
	
	/**
	 * This method detects whether this category can unify with the given category. Neither of the two
	 * categories are changed.
	 * 
	 * @param category The category for the unification check.
	 * @return true if the two categories can unify.
	 */
	public boolean canUnify(Category category) {
		if (!isSimilar(category)) return false;
		Category thisC = deepCopy();
		Category otherC = category.deepCopy();
		try {
			thisC.tryToUnify(otherC);
		} catch (UnificationFailedException ex) {
			return false;
		}
		return true;
	}
	
	/**
	 * Skolemizes the feature values of this category.
	 */
	public void skolemize() {
		featureMap.skolemize();
	}
	
	/**
	 * Creates a deep copy of this category.
	 * 
	 * @return A deep copy.
	 */
	public Category deepCopy() {
		return deepCopy(new HashMap<Integer, StringEntity>());
	}
	
	/**
	 * Creates a deep copy of this category using the given string entities. This method is
	 * usually called form another deepCopy-method.
	 * 
	 * @param entities The string entities to be used.
	 * @return A deep copy.
	 */
	Category deepCopy(HashMap<Integer, StringEntity> entities) {
		Category c;
		if (this instanceof Terminal) {
			c = new Terminal(name);
		} else {
			c = new Nonterminal(name);
		}
		c.featureMap = featureMap.deepCopy(entities);
		return c;
	}
	
	/**
	 * This methods checks whether two categories are similar. Two categories are similar if and
	 * only if the categories have the same name and no feature with the same name is present in
	 * both categories with values that do not unify locally.
	 * 
	 * @param c The category for which similarity with this category should be checked.
	 * @return true if the two categories are similar.
	 */
	public boolean isSimilar(Category c) {
		if (!name.equals(c.name)) return false;
		return featureMap.isSimilar(c.featureMap);
	}
	
	/**
	 * This method returns true if this category subsumes (i.e. is more general than) the given category,
	 * or false otherwise.
	 * 
	 * @param category The category for which it is checked whether this category subsumes it.
	 * @return true if this category subsumes the given category.
	 */
	public boolean subsumes(Category category) {
		if (!isSimilar(category)) return false;
		
		// Both categories are copied to keep the existing categories untouched:
		Category category1C = deepCopy();
		Category category2C = category.deepCopy();
		
		// Category 1 subsumes category 2 iff 1 unifies with 2 after the skolemization of 2.
		category2C.skolemize();
		try {
			category1C.tryToUnify(category2C);
			return true;
		} catch (UnificationFailedException ex) {
			return false;
		}
	}
	
	public boolean equals(Object obj) {
		if (!(obj instanceof Category)) return false;
		return toString().equals(obj.toString());
	}
	
	public String toString() {
		return name + featureMap.toString();
	}

}
