// This file is part of the Attempto Java Packages.
// Copyright 2008-2009, Attempto Group, University of Zurich (see http://attempto.ifi.uzh.ch).
//
// The Attempto Java Packages is free software: you can redistribute it and/or modify it under the
// terms of the GNU Lesser General Public License as published by the Free Software Foundation,
// either version 3 of the License, or (at your option) any later version.
//
// The Attempto Java Packages is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE. See the GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License along with the Attempto
// Java Packages. If not, see http://www.gnu.org/licenses/.

package ch.uzh.ifi.attempto.chartparser;

import java.util.ArrayList;
import java.util.HashMap;

/**
 * This class represents a chart (in the sense of a chart for a chart parser). A chart basically consists
 * of a set of edges plus operations to add, remove, and retrieve edges. The edges stored by a chart represent
 * (partial) analyses of certain ranges of a text.
 * 
 * @author Tobias Kuhn
 */
class Chart {
	
	// All edges are contained by this list:
	private ArrayList<Edge> edges = new ArrayList<Edge>();
	
	// This hashtable indexes the edges by their end positions:
	private HashMap<Integer, ArrayList<Edge>> edgesByEndPos = new HashMap<Integer, ArrayList<Edge>>();
	
	// This hashtable indexes the edges by their end positions and active categories:
	private HashMap<Integer, HashMap<String, ArrayList<Edge>>> edgesByEndPosAndActCat =
		new HashMap<Integer, HashMap<String, ArrayList<Edge>>>();
	
	/**
	 * Creates an empty chart.
	 */
	public Chart() {
	}
	
	/**
	 * Adds the given edge to the chart. If there is already an edge in the chart that subsumes the
	 * given edge then the chart remains unchanged and false is returned.
	 * 
	 * @param edge The edge to be added to the chart.
	 * @return true if the edge was added to the chart or false if there is a subsuming edge in the chart.
	 */
	public boolean addEdge(Edge edge) {
		// Get the list of all existing edges with the same end position:
		ArrayList<Edge> l = getEdgesByEndPos(edge.getEndPos());
		
		// Check whether the edge is new (i.e. not subsumed by an existing edge):
		boolean isNewEdge = true;
		for (Edge e : l) {
			if (e.subsumes(edge) && edge.subsumes(e)) {
				e.addLinksFrom(edge);
				isNewEdge = false;
				break;
			}
		}
		
		if (isNewEdge) {
			// Add the edge to the chart:
			edges.add(edge);
			
			// Update all indexes:
			l.add(edge);
			String n = ( edge.isActive() ? edge.getNextActive().getName() : null );
			getEdgesByEndPosAndActCat(edge.getEndPos(), n).add(edge);
		}
		
		return isNewEdge;
	}
	
	/**
	 * Returns the size of the chart, i.e. the number of edges.
	 * 
	 * @return The size of the chart.
	 */
	public int getSize() {
		return edges.size();
	}
	
	/**
	 * Returns a list of all edges with the given end position. The returned list in an internal list
	 * that should never be changed from outside.
	 * 
	 * @param endPos The end position of the edges to be returned.
	 * @return A list of all edges with the given end position.
	 */
	public ArrayList<Edge> getEdgesByEndPos(int endPos) {
		ArrayList<Edge> l = edgesByEndPos.get(endPos);
		if (l == null) {
			l = new ArrayList<Edge>();
			edgesByEndPos.put(endPos, l);
		}
		return l;
	}
	
	/**
	 * Returns a list of all edges with the given end position and active category name. The returned
	 * list is an interal list that should never be changed from outside.
	 * 
	 * @param endPos The end position of the edges to be returned.
	 * @param actCatName The active category name of the edges to be returned.
	 * @return A list of all edges with the given end position and active category name.
	 */
	public ArrayList<Edge> getEdgesByEndPosAndActCat(int endPos, String actCatName) {
		HashMap<String, ArrayList<Edge>> h = edgesByEndPosAndActCat.get(endPos);
		if (h == null) {
			h = new HashMap<String, ArrayList<Edge>>();
			edgesByEndPosAndActCat.put(endPos, h);
		}
		ArrayList<Edge> l = h.get(actCatName);
		if (l == null) {
			l = new ArrayList<Edge>();
			h.put(actCatName, l);
		}
		return l;
	}
	
	/**
	 * Removes all edges with the given end position from the chart.
	 * 
	 * @param endPos The end position of the edges to be removed.
	 */
	public void removeEdgesWithEndPos(int endPos) {
		ArrayList<Edge> l = edgesByEndPos.get(endPos);
		if (l == null) return;
		for (Edge e : l) {
			edges.remove(e);
		}
		l.clear();
		
		HashMap<String, ArrayList<Edge>> h = edgesByEndPosAndActCat.get(endPos);
		if (h == null) return;
		h.clear();
	}
	
	/**
	 * Removes all edges from the chart.
	 */
	public void clear() {
		edges.clear();
		edgesByEndPos.clear();
		edgesByEndPosAndActCat.clear();
	}
	
	public String toString() {
		String s = "";
		for (Edge e : edges) {
			s += e + "\n";
		}
		return s;
	}

}
