// This file is part of the Attempto Java Packages.
// Copyright 2008-2009, Attempto Group, University of Zurich (see http://attempto.ifi.uzh.ch).
//
// The Attempto Java Packages is free software: you can redistribute it and/or modify it under the
// terms of the GNU Lesser General Public License as published by the Free Software Foundation,
// either version 3 of the License, or (at your option) any later version.
//
// The Attempto Java Packages is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE. See the GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License along with the Attempto
// Java Packages. If not, see http://www.gnu.org/licenses/.

package ch.uzh.ifi.attempto.chartparser;

import java.util.ArrayList;
import java.util.List;

/**
 * This is a chart parser (concretely an Earley parser) fully implemented in Java. However, there is a
 * Prolog-based format called "Codeco" that can be transformed into Java
 * (at compile time).
 * 
 * @author Tobias Kuhn
 * @see Grammar
 */
public class ChartParser {
	
	private final Grammar grammar;
	private final Chart chart = new Chart();
	private final ArrayList<Terminal[]> tokens = new ArrayList<Terminal[]>();
	private boolean debug;
	
	/**
	 * Creates a new chart parser for the given grammar. The grammar must not be changed afterwards.
	 * 
	 * @param grammar The grammar to be used by the chart parser.
	 */
	public ChartParser(Grammar grammar) {
		this.grammar = grammar;
		init();
		completeAndPredict();
	}
	
	/**
	 * This method can be used to switch on/off debug mode (default is off). In debug mode, messages about the actions
	 * of the chart parser are printed onto the standard error device.
	 * 
	 * @param debug true to switch on debug mode or false to switch it off.
	 */
	public void debug(boolean debug) {
		this.debug = debug;
	}
	
	/**
	 * Adds the token to the token sequence and makes one more parsing step to process it. If several tokens
	 * are given then these tokens are treated in a parallel way representing a token that can have multiple
	 * forms.
	 * 
	 * @param token The token to be added to the token sequence.
	 */
	public void addToken(Terminal... token) {
		tokens.add(token);
		for (Terminal t : token) {
			t.setFeature("id", tokens.size() + "");
			Edge edge = new Edge(tokens.size()-1, tokens.size(), t, true);
			chart.addEdge(edge);
			if (debug) log("SCANNER: " + edge + "\n");
		}
		completeAndPredict();
		//if (debug) log("CHART:");
		//if (debug) log(chart);
	}
	
	/**
	 * Adds the tokens to the token sequence and processes them.
	 * 
	 * @param tokens The tokens to be added to the token sequence.
	 */
	public void addTokens(List<Terminal> tokens) {
		for (Terminal t : tokens) {
			addToken(t);
		}
	}
	
	/**
	 * Removes the last token and reverts the last parsing step.
	 */
	public void removeToken() {
		chart.removeEdgesWithEndPos(tokens.size());
		tokens.remove(tokens.size()-1);
	}
	
	/**
	 * Removes all tokens in the current token sequence and resets the chart.
	 */
	public void removeAllTokens() {
		tokens.clear();
		chart.clear();
		init();
		completeAndPredict();
	}
	
	/**
	 * Returns the current token sequence.
	 * 
	 * @return The current token sequence.
	 */
	public List<Terminal[]> getTokens() {
		return new ArrayList<Terminal[]>(tokens);
	}

	/**
	 * This method looks ahead and returns the restrictions at least one of which the next token
	 * has to fulfill.
	 * 
	 * @return The restrictions.
	 */
	public List<Restriction> getNextTokenRestrictions() {
		if (debug) log("LOOKING FORWARD:");
		
		ArrayList<Edge> edges = new ArrayList<Edge>();
		for (Edge e : chart.getEdgesByEndPos(tokens.size())) {
			if (!e.isActive()) continue;
			if (!(e.getNextActive() instanceof Terminal)) continue;
			
			if (e.getApplicableBackref() != null) {
				// For edges with backwards references, the possible bindings have to be performed:
				for (int i = 0 ; i < e.getExternalAnteList().length ; i++) {
					Edge eC = e.deepCopy();
					try {
						eC.getExternalAnteList()[i].unify(eC.getApplicableBackref());
						edges.add(eC);
					} catch (UnificationFailedException ex) {}
				}
			} else {
				edges.add(e);
			}
		}

		ArrayList<Restriction> restrictions = new ArrayList<Restriction>();
		for (Edge e : edges) {
			Terminal t = (Terminal) e.getNextActive();
			List<Terminal> exceptions = new ArrayList<Terminal>();
			if (e.getApplicableNegbackref() != null) {
				// Edges with negative backwards references lead to exceptions:
				for (int i = 0 ; i < e.getExternalAnteList().length ; i++) {
					Edge eC = e.deepCopy();
					try {
						eC.getExternalAnteList()[i].unify(eC.getApplicableNegbackref());
						exceptions.add((Terminal) eC.getNextActive());
					} catch (UnificationFailedException ex) {}
				}
			}
			restrictions.add(new Restriction((Terminal) t.deepCopy(), exceptions));
			if (debug) log(" " + t);
		}
		if (debug) log("\n");
		
		return restrictions;
	}
	
	private void init() {
		Category startCat = grammar.getStartCategory();
		for (Rule rule : grammar.getRulesByHeadName(startCat.getName())) {
			try {
				if (!startCat.isSimilar(rule.getHead())) continue;
				Nonterminal categoryC = (Nonterminal) startCat.deepCopy();
				Rule ruleC = rule.deepCopy();
				categoryC.unify(ruleC.getHead());
				Edge edge = new Edge(0, 0, ruleC.getHead(), ruleC.getBody(), ruleC.isAccessible());
				chart.addEdge(edge);
				if (debug) log("INIT: " + rule + "  --->  " + edge + "\n");
			} catch (UnificationFailedException ex) {
				continue;
			}
		}
	}
	
	private void completeAndPredict() {
		// Process the epsilon rules:
		for (Rule rule : grammar.getEpsilonRules()) {
			// Epsilon edges get no antecedent lists (they cannot have references):
			Edge edge = new Edge(tokens.size(), tokens.size(), rule.getHead(), rule.isAccessible());
			boolean isNewEdge = chart.addEdge(edge);
			if (debug) log("EDGE FOR EPSILON RULE (" + (isNewEdge ? "NEW" : "KNOWN") + "): " + edge + "\n");
		}
		
		// Run completion/predition until neither of them generates a new edge:
		int s = 0;
		while (chart.getSize() > s) {
			s = chart.getSize();
			complete();
			resolve();
			predict();
		}
	}
	
	private void predict() {
		ArrayList<Edge> l = chart.getEdgesByEndPos(tokens.size());
		for (int i = 0 ; i < l.size() ; i++) {
			// During this loop, elements might be added to the end of the list l.
			
			Edge existingEdge = l.get(i);
			Category category = existingEdge.getNextActive();
			if (category == null) continue;
			if (debug) log("PREDICTION FOR CATEGORY: " + category + "\n");
			
			for (Rule rule : grammar.getRulesByHeadName(category.getName())) {
				if (rule.hasEmptyBody()) continue;
				
				try {
					if (!category.isSimilar(rule.getHead())) continue;
					Category categoryC = category.deepCopy();
					Rule ruleC = rule.deepCopy();
					categoryC.unify(ruleC.getHead());
					int p = tokens.size();
					Edge edge = new Edge(p, ruleC, existingEdge.getCombinedAnteList());
					boolean isNewEdge = chart.addEdge(edge);
					if (debug) log("PREDICT (" + (isNewEdge ? "NEW" : "KNOWN") + "): " + rule + "  --->  " + edge + "\n");
				} catch (UnificationFailedException ex) {
					continue;
				}
			}
		}
	}
	
	private void complete() {
		ArrayList<Edge> l1 = chart.getEdgesByEndPosAndActCat(tokens.size(), null);
		for (int i1 = 0 ; i1 < l1.size() ; i1++) {
			// During this loop, elements might be added to the end of the list l1.
			
			Edge passiveEdge = l1.get(i1);
			if (debug) log("COMPLETION FOR EDGE: " + passiveEdge + "\n");
			
			ArrayList<Edge> l2 = chart.getEdgesByEndPosAndActCat(passiveEdge.getStartPos(), passiveEdge.getHead().getName());
			for (int i2 = 0 ; i2 < l2.size() ; i2++) {
				// During this loop, elements might be added to the end of the list l2.
				
				Edge edge = l2.get(i2);
				if (!edge.isActive()) continue;
				
				try {
					if (!passiveEdge.getHead().isSimilar(edge.getNextActive())) continue;
					Edge passiveEdgeC = passiveEdge.deepCopy();
					Edge edgeC = edge.deepCopy();
					passiveEdgeC.getHead().unify(edgeC.getNextActive());
					if (!passiveEdge.hasEmptyBody()) {
						// Antecedent lists have to match (except for the case of epsilon edges):
						FeatureMap[] al1 = edgeC.getCombinedAnteList();
						FeatureMap[] al2 = passiveEdgeC.getExternalAnteList();
						if (al1.length != al2.length) throw new UnificationFailedException();
						for (int i = 0 ; i < al1.length ; i++) {
							al1[i].unify(al2[i]);
						}
					}
					edgeC.step(tokens.size(), passiveEdgeC);
					boolean isNewEdge = chart.addEdge(edgeC);
					if (debug) log("COMPLETE (" + (isNewEdge ? "NEW" : "KNOWN") + "): " + edge + "  --->  " + edgeC + "\n");
				} catch (UnificationFailedException ex) {
					continue;
				}
			}
		}
	}
	
	private void resolve() {
		ArrayList<Edge> l1 = chart.getEdgesByEndPos(tokens.size());
		for (int i1 = 0 ; i1 < l1.size() ; i1++) {
			// During this loop, elements might be added to the end of the list l1.
			
			Edge edge = l1.get(i1);
			if (!edge.isActive()) continue;
			
			String n = edge.getNextActive().getName();
			Edge edgeC = null;
			if (n.equals(">>>")) {
				edgeC = edge.deepCopy();
				edgeC.addAntecedents(edgeC.getNextActive().getFeatureMap());
			} else if (n.equals("<<<")) {
				FeatureMap[] ante = edge.getCombinedAnteList();
				for (int i = ante.length-1 ; i >= 0 ; i--) {
					if (ante[i].canUnify(edge.getNextActive().getFeatureMap())) {
						try {
							edgeC = edge.deepCopy();
							edgeC.getExternalAnteList()[i].unify(edgeC.getNextActive().getFeatureMap());
							break;
						} catch (UnificationFailedException ex) {
							edgeC = null;
						}
					}
				}
			} else if (n.equals("</<")) {
				edgeC = edge.deepCopy();
				for (FeatureMap f : edge.getCombinedAnteList()) {
					// TODO: This check is probably not sufficient:
					if (f.canUnify(edge.getNextActive().getFeatureMap())) {
						edgeC = null;
						break;
					}
				}
			} else {
				continue;
			}
			
			if (edgeC == null) {
				if (debug) log("CANNOT RESOLVE: " + edge + "\n");
			} else {
				edgeC.step();
				boolean isNewEdge = chart.addEdge(edgeC);
				if (debug) log("RESOLVE (" + (isNewEdge ? "NEW" : "KNOWN") + "): " + edge + "  --->  " + edgeC + "\n");
			}
		}
	}
	
	private void log(String text) {
		System.err.print(text);
	}

}
