// This file is part of the Attempto Java Packages.
// Copyright 2008-2009, Attempto Group, University of Zurich (see http://attempto.ifi.uzh.ch).
//
// The Attempto Java Packages is free software: you can redistribute it and/or modify it under the
// terms of the GNU Lesser General Public License as published by the Free Software Foundation,
// either version 3 of the License, or (at your option) any later version.
//
// The Attempto Java Packages is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE. See the GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License along with the Attempto
// Java Packages. If not, see http://www.gnu.org/licenses/.

package ch.uzh.ifi.attempto.chartparser;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Set;

/**
 * This class represents an edge of a chart to be used by a chart parser. Edges are derived from grammar
 * rules. An edge contains the head and body categories of the rule it is derived from. In addition, an
 * edge has a start position and an end position that define the range of the text that is covered.
 * Furthermore, every edge has a progress value that records how many of the body categories are parsed.
 * Edges that have unparsed categories are called "active" whereas edges for which all body categories are
 * parsed are called "passive". In order to capture accessibility constraints for anaphoric
 * references, edges are either accessible or inaccessible, like rules. Additionally, edges contain two lists that
 * store the feature maps of the accessible antecedents. The first list stores the antecedents from outside the
 * range of the edge (the external antecedent list) and the second list stores the antecedents from inside the
 * range of the edge (the internal antecedent list).
 * 
 * @author Tobias Kuhn
 */
class Edge {
	
	private int startPos;
	private int endPos;
	private Category head;
	private FeatureMap[] externalAnteList = new FeatureMap[0];
	private FeatureMap[] internalAnteList = new FeatureMap[0];
	private boolean accessible;
	private Category[] body;
	private int progress = 0;
	private HashSet<Edge> links = new HashSet<Edge>();
	
	private Edge() {
	}
	
	/**
	 * Creates a new edge on the basis of a rule.
	 * 
	 * @param pos Start and end position.
	 * @param rule The rule that is the basis for the creation of the edge.
	 * @param externalAnteList The list of external antecedents.
	 */
	public Edge(int pos, Rule rule, FeatureMap[] externalAnteList) {
		this.startPos = pos;
		this.endPos = pos;
		this.head = rule.getHead();
		this.body = rule.getBody();
		this.externalAnteList = (FeatureMap[]) externalAnteList.clone();
		this.accessible = rule.isAccessible();
	}
	
	/**
	 * Creates a new edge with a progress value of 0.
	 * 
	 * @param startPos The start position.
	 * @param endPos The end position.
	 * @param head The head category.
	 * @param body The body categories.
	 * @param accessible true for accessible edges.
	 */
	public Edge(int startPos, int endPos, Category head, Category[] body, boolean accessible) {
		this.accessible = accessible;
		this.startPos = startPos;
		this.endPos = endPos;
		this.head = head;
		this.body = (Category[]) body.clone();
	}
	
	/**
	 * Creates a new edge with no body categories.
	 * 
	 * @param startPos The start position.
	 * @param endPos The end position.
	 * @param head The head category.
	 * @param accessible true for accessible edges.
	 */
	public Edge(int startPos, int endPos, Category head, boolean accessible) {
		this.accessible = accessible;
		this.startPos = startPos;
		this.endPos = endPos;
		this.head = head;
		this.body = new Category[0];
	}
	
	/**
	 * Returns the start position of this edge.
	 * 
	 * @return The start position.
	 */
	public int getStartPos() {
		return startPos;
	}
	
	/**
	 * Returns the end position of this edge.
	 * 
	 * @return The end position.
	 */
	public int getEndPos() {
		return endPos;
	}
	
	/**
	 * Returns the head category of this edge.
	 * 
	 * @return The head category.
	 */
	public Category getHead() {
		return head;
	}
	
	/**
	 * Returns the body categories of this edge.
	 * 
	 * @return The body categories.
	 */
	public Category[] getBody() {
		return body;
	}
	
	/**
	 * Returns the progress value of this edge.
	 * 
	 * @return The progress value.
	 */
	public int getProgress() {
		return progress;
	}
	
	/**
	 * Returns the list of external antecedents.
	 * 
	 * @return The external antecedents.
	 */
	public FeatureMap[] getExternalAnteList() {
		return externalAnteList;
	}
	
	/**
	 * Returns the list of internal antecedents.
	 * 
	 * @return The internal antecedents.
	 */
	public FeatureMap[] getInternalAnteList() {
		return internalAnteList;
	}
	
	/**
	 * Returns a list that is a concatenation of the lists of external and internal antecedents.
	 * 
	 * @return A combined antecedents list.
	 */
	public FeatureMap[] getCombinedAnteList() {
		FeatureMap[] l = new FeatureMap[externalAnteList.length + internalAnteList.length];
		for (int i = 0 ; i < externalAnteList.length ; i++) {
			l[i] = externalAnteList[i];
		}
		for (int i = 0 ; i < internalAnteList.length ; i++) {
			l[externalAnteList.length + i] = internalAnteList[i];
		}
		return l;
	}
	
	/**
	 * Returns all edges that have been used to parse the parsed body categories. This is
	 * empty if the progress value is 0.
	 * 
	 * @return A set of edges that have been used for the (partial) parsing of this edge.
	 */
	public Set<Edge> getLinks() {
		return links;
	}
	
	/**
	 * Returns true if this edge is active, i.e. has unparsed body categories.
	 * 
	 * @return true if this edge is active.
	 */
	public boolean isActive() {
		return body.length > progress;
	}
	
	/**
	 * Returns true if this edge is accessible, i.e. is derived from an accessible rule.
	 * 
	 * @return true if this edge is accessible.
	 */
	public boolean isAccessible() {
		return accessible;
	}
	
	/**
	 * Returns true if the edge has an empty body, i.e. it is an epsilon edge.
	 * 
	 * @return true if this is an epsilon edge.
	 */
	public boolean hasEmptyBody() {
		return body.length == 0;
	}
	
	/**
	 * Returns the first active category, i.e. the first body category that has not yet been
	 * parsed. null is returned for inactive edges.
	 * 
	 * @return The first active category.
	 */
	public Category getNextActive() {
		if (!isActive()) return null;
		return body[progress];
	}
	
	/**
	 * Moves the parsing progress forward to the given position. The given edge is stored as a link.
	 * No checking is done within this method whether this operation is actually possible according
	 * to the rules of chart parsing.
	 * 
	 * @param pos The end position to move forward to.
	 * @param edge The edge that is used for moving forward.
	 */
	public void step(int pos, Edge edge) {
		if (!isActive()) {
			throw new RuntimeException("Passive edge");
		}
		links.add(edge);
		progress++;
		endPos = pos;
		if (edge.isAccessible()) {
			addAntecedents(edge.getInternalAnteList());
		}
	}
	
	/**
	 * Moves the parsing progress one step forward while keeping the same end position.
	 * No checking is done within this method whether this operation is actually possible according
	 * to the rules of chart parsing.
	 */
	public void step() {
		if (!isActive()) {
			throw new RuntimeException("Passive edge");
		}
		progress++;
	}
	
	/**
	 * Adds the links for the given edge to this edge.
	 * 
	 * @param edge The edge whose links should be added.
	 */
	public void addLinksFrom(Edge edge) {
		links = new HashSet<Edge>(edge.links);
	}
	
	/**
	 * Returns the feature map of a backwards reference that is applicable for prediction. Null is
	 * returned if there is no such reference.
	 * 
	 * @return The feature map of an applicable backwards reference.
	 */
	public FeatureMap getApplicableBackref() {
		if (!(body[progress] instanceof Terminal)) return null;
		for (int i = progress + 1 ; i < body.length ; i++) {
			if (body[i] instanceof Terminal) continue;
			if (body[i].getName().equals("<<<")) return body[i].getFeatureMap();
			return null;
		}
		return null;
	}
	
	/**
	 * Returns the feature map of a negative backwards reference that is applicable for prediction.
	 * Null is returned if there is no such reference.
	 * 
	 * @return The feature map of an applicable negative backwards reference.
	 */
	public FeatureMap getApplicableNegbackref() {
		int l = body.length;
		int p = progress;
		if (p + 1 < l && body[p] instanceof Terminal && body[p + 1].getName().equals("</<")) {
			return body[p + 1].getFeatureMap();
		}
		return null;
	}
	
	/**
	 * Adds the feature maps to the list of internal antecedents.
	 * 
	 * @param newAnteList The feature maps of internal antecedents.
	 */
	public void addAntecedents(FeatureMap... newAnteList) {
		FeatureMap[] oldInternalList = internalAnteList;
		int l1 = oldInternalList.length;
		int l2 = newAnteList.length;
		internalAnteList = new FeatureMap[l1 + l2];
		for (int i = 0 ; i < l1 ; i++) {
			internalAnteList[i] = oldInternalList[i];
		}
		for (int i = 0 ; i < l2 ; i++) {
			internalAnteList[l1 + i] = newAnteList[i];
		}
	}
	
	/**
	 * Skolemizes the variables of this edge.
	 */
	public void skolemize() {
		head.skolemize();
		for (FeatureMap c : externalAnteList) c.skolemize();
		for (FeatureMap c : internalAnteList) c.skolemize();
		for (Category c : body) c.skolemize();
	}
	
	/**
	 * Unifies this edge with the given edge. If the unification fails, a
	 * UnificationFailedException is thrown. In this case, the two edges remain partly unified,
	 * i.e. no backtracking is done. Thus, this operation should be perfomed only if it is certain that
	 * the unification succeeds, or if the operation is performed on copies of objects that are not
	 * used anymore afterwards.
	 * 
	 * @param edge The edge to be unified with this edge.
	 * @throws UnificationFailedException If unification fails.
	 */
	public void unify(Edge edge) throws UnificationFailedException {
		if (accessible != edge.accessible) throw new UnificationFailedException();
		if (body.length != edge.body.length) throw new UnificationFailedException();
		if (progress != edge.progress) throw new UnificationFailedException();
		if (startPos != edge.startPos) throw new UnificationFailedException();
		if (endPos != edge.endPos) throw new UnificationFailedException();
		if (externalAnteList.length != edge.externalAnteList.length) throw new UnificationFailedException();
		if (internalAnteList.length != edge.internalAnteList.length) throw new UnificationFailedException();
		
		head.unify(edge.head);
		for (int i = 0 ; i < externalAnteList.length ; i++) {
			externalAnteList[i].unify(edge.externalAnteList[i]);
		}
		for (int i = 0 ; i < internalAnteList.length ; i++) {
			internalAnteList[i].unify(edge.internalAnteList[i]);
		}
		for (int i = 0 ; i < body.length ; i++) {
			body[i].unify(edge.body[i]);
		}
	}
	
	/**
	 * Tries to unify this edge with another edge. If unification is not possible, an exception
	 * is thrown. In the case unification would be possible, the unification is not performed completely.
	 * In any case the two edges remain in an unconsistent state afterwards. Thus, this operation should
	 * be performed only on copies of objects that are not used anymore afterwards.
	 * 
	 * @param edge The edge to be unified with this edge.
	 * @throws UnificationFailedException If unification fails.
	 */
	public void tryToUnify(Edge edge) throws UnificationFailedException {
		if (accessible != edge.accessible) throw new UnificationFailedException();
		if (body.length != edge.body.length) throw new UnificationFailedException();
		if (progress != edge.progress) throw new UnificationFailedException();
		if (startPos != edge.startPos) throw new UnificationFailedException();
		if (endPos != edge.endPos) throw new UnificationFailedException();
		if (externalAnteList.length != edge.externalAnteList.length) throw new UnificationFailedException();
		if (internalAnteList.length != edge.internalAnteList.length) throw new UnificationFailedException();
		if (externalAnteList.length != edge.externalAnteList.length) throw new UnificationFailedException();
		if (internalAnteList.length != edge.internalAnteList.length) throw new UnificationFailedException();
		
		head.tryToUnify(edge.head);
		for (int i = 0 ; i < externalAnteList.length ; i++) {
			externalAnteList[i].tryToUnify(edge.externalAnteList[i]);
		}
		for (int i = 0 ; i < internalAnteList.length ; i++) {
			internalAnteList[i].tryToUnify(edge.internalAnteList[i]);
		}
		for (int i = 0 ; i < body.length ; i++) {
			body[i].tryToUnify(edge.body[i]);
		}
	}
	
	/**
	 * This methods checks whether two edges are similar. Two edges are similar if and only if
	 * the replacement of every variable occurence by a new singleton variable (in both edges)
	 * would make them equal (apart from variable names). Edges that are not similar can never
	 * unify. Similar edges may or may not unify.
	 * 
	 * @param edge The edge for which similarity with this edge should be checked.
	 * @return true if the two edges are similar.
	 */
	public boolean isSimilar(Edge edge) {
		if (accessible != edge.accessible) return false;
		if (startPos != edge.startPos) return false;
		if (endPos != edge.endPos) return false;
		if (!head.isSimilar(edge.head)) return false;
		if (body.length != edge.body.length) return false;
		if (progress != edge.progress) return false;
		if (externalAnteList.length != edge.externalAnteList.length) return false;
		if (internalAnteList.length != edge.internalAnteList.length) return false;
		for (int i = 0 ; i < externalAnteList.length ; i++) {
			if (!externalAnteList[i].isSimilar(edge.externalAnteList[i])) return false;
		}
		for (int i = 0 ; i < internalAnteList.length ; i++) {
			if (!internalAnteList[i].isSimilar(edge.internalAnteList[i])) return false;
		}
		for (int i = 0 ; i < body.length ; i++) {
			if (!body[i].isSimilar(edge.body[i])) return false;
		}
		return true;
	}
	
	/**
	 * This method returns true if this edge subsumes (i.e. is more general than) the given edge,
	 * or false otherwise.
	 * 
	 * @param edge The edge for which it is checked whether this edge subsumes it.
	 * @return true if this edge subsumes the given edge.
	 */
	public boolean subsumes(Edge edge) {
		if (!isSimilar(edge)) return false;

		// Both edges are copied to keep the existing edges untouched:
		Edge edge1C = deepCopy();
		Edge edge2C = edge.deepCopy();
		
		// Edge 1 subsumes edge 2 iff 1 unifies with 2 after the skolemization of 2.
		edge2C.skolemize();
		try {
			edge1C.tryToUnify(edge2C);
			return true;
		} catch (UnificationFailedException ex) {
			return false;
		}
	}
	
	/**
	 * Creates a deep copy of this edge.
	 * 
	 * @return A deep copy.
	 */
	public Edge deepCopy() {
		return deepCopy(new HashMap<Integer, StringEntity>());
	}
	
	/**
	 * Creates a deep copy of this edge using the given string entities. This method is
	 * usually called form another deepCopy-method.
	 * 
	 * @param entities The string entities to be used.
	 * @return A deep copy.
	 */
	Edge deepCopy(HashMap<Integer, StringEntity> entities) {
		Edge edgeC = new Edge();
		edgeC.startPos = startPos;
		edgeC.endPos = endPos;
		edgeC.progress = progress;
		edgeC.accessible = accessible;
		edgeC.head = head.deepCopy(entities);
		edgeC.body = new Category[body.length];
		for (int i=0 ; i < body.length ; i++) {
			edgeC.body[i] = body[i].deepCopy(entities);
		}
		edgeC.externalAnteList = new FeatureMap[externalAnteList.length];
		for (int i=0 ; i < externalAnteList.length ; i++) {
			edgeC.externalAnteList[i] = externalAnteList[i].deepCopy(entities);
		}
		edgeC.internalAnteList = new FeatureMap[internalAnteList.length];
		for (int i=0 ; i < internalAnteList.length ; i++) {
			edgeC.internalAnteList[i] = internalAnteList[i].deepCopy(entities);
		}
		edgeC.addLinksFrom(this);
		return edgeC;
	}
	
	public String toString() {
		String s = "<" + startPos + "," + endPos + "> " + head;
		s += " ";
		for (FeatureMap f : externalAnteList) {
			s += f;
		}
		if (accessible) {
			s += "=";
		} else {
			s += "~";
		}
		for (FeatureMap f : internalAnteList) {
			s += f;
		}
		s += ">";
		for (int i=0 ; i < body.length ; i++) {
			if (i == progress) s += " .";
			s += " " + body[i];
		}
		if (!isActive()) s += " .";
		return s;
	}

}
