// This file is part of the Attempto Java Packages.
// Copyright 2008-2009, Attempto Group, University of Zurich (see http://attempto.ifi.uzh.ch).
//
// The Attempto Java Packages is free software: you can redistribute it and/or modify it under the
// terms of the GNU Lesser General Public License as published by the Free Software Foundation,
// either version 3 of the License, or (at your option) any later version.
//
// The Attempto Java Packages is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE. See the GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License along with the Attempto
// Java Packages. If not, see http://www.gnu.org/licenses/.

package ch.uzh.ifi.attempto.chartparser;

import java.util.ArrayList;
import java.util.List;

/**
 * This class describes a restriction for the next token of a given token sequence. It
 * contains of a category and of a number of exceptions which are again categories.
 * In order to fulfill the restriction, the category of a token must be subsumed by the
 * category of the restriction while not being subsumed by any of the exceptions. The
 * exceptions come from negative backwards references in the grammar.
 * 
 * @author Tobias Kuhn
 */
public class Restriction {
	
	private Terminal category;
	private List<Terminal> exceptions;
	
	/**
	 * Creates a new restrictions with the given category and no exceptions.
	 * 
	 * @param category The category.
	 */
	Restriction(Terminal category) {
		this.category = category;
		exceptions = new ArrayList<Terminal>();
	}
	
	/**
	 * Creates a new restriction with the given category and exceptions.
	 * 
	 * @param category The category.
	 * @param exceptions A list of exceptions.
	 */
	Restriction(Terminal category, List<Terminal> exceptions) {
		this.category = category;
		this.exceptions = exceptions;
	}
	
	/**
	 * Returns the category of this restriction.
	 * 
	 * @return The category.
	 */
	public Terminal getCategory() {
		return category;
	}
	
	/**
	 * Returns the list of exceptions of this restriction.
	 * 
	 * @return The list of exceptions.
	 */
	public List<Terminal> getExceptions() {
		return exceptions;
	}
	
	/**
	 * Returns true if the given category fulfills this restriction.
	 * 
	 * @param t The category to be tested.
	 * @return true if the category fulfills this restriction.
	 */
	public boolean isFulfilledBy(Terminal t) {
		if (!category.subsumes(t)) return false;
		for (Terminal r : exceptions) {
			if (r.subsumes(t)) return false;
		}
		return true;
	}

}
