// This file is part of the Attempto Java Packages.
// Copyright 2008-2009, Attempto Group, University of Zurich (see http://attempto.ifi.uzh.ch).
//
// The Attempto Java Packages is free software: you can redistribute it and/or modify it under the
// terms of the GNU Lesser General Public License as published by the Free Software Foundation,
// either version 3 of the License, or (at your option) any later version.
//
// The Attempto Java Packages is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE. See the GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License along with the Attempto
// Java Packages. If not, see http://www.gnu.org/licenses/.

package ch.uzh.ifi.attempto.preditor;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

import ch.uzh.ifi.attempto.chartparser.ChartParser;
import ch.uzh.ifi.attempto.chartparser.Restriction;
import ch.uzh.ifi.attempto.chartparser.Terminal;
import ch.uzh.ifi.attempto.preditor.text.TextElement;

/**
 * This abstract class is used for the predictive editor to create menus on the basis of categories of the grammar.
 *  
 * @author Tobias Kuhn
 */
public abstract class MenuCreator {
	
	private HashMap<String, MenuBlockContent> contentsMap = new HashMap<String, MenuBlockContent>();
	private List<MenuBlockContent> contentsList = new ArrayList<MenuBlockContent>();
	private Restriction restriction;
	private int id;
	
	/**
	 * Initializes a new menu creator object.
	 */
	public MenuCreator() {
	}
	
	/**
	 * An implementation of this abstract method should add menu items that comply with the given restriction.
	 * This should be done by calling {@link #addMenuItem addMenuItem} and/or {@link #addMenuEntry addMenuEntry}.
	 * 
	 * @param restriction The restriction for the menu items.
	 */
	protected abstract void addMenuItems(Restriction restriction);
	
	/**
	 * This abstract method is called when the predictive editor is refreshed (before the first call of
	 * {@link #addMenuItems addMenuItems}). This is a good place for calling {@link #prepareMenuBlock(String) prepareMenuBlock}.
	 */
	public abstract void initMenuCreation();
	
	/**
	 * Prepares a new menu block (if there is no such menu block already with the same name). The menu blocks are
	 * shown in the predictive editor in the same order as they have been prepared.
	 * 
	 * @param menuBlockName The name of the menu block.
	 */
	public void prepareMenuBlock(String menuBlockName) {
		prepareMenuBlock(menuBlockName, false);
	}

	/**
	 * Prepares a new menu block (if there is no such menu block already with the same name). The second parameter
	 * defines whether the items of the menu block should be sorted. The menu blocks are
	 * shown in the predictive editor in the same order as they have been prepared.
	 * 
	 * @param menuBlockName The name of the menu block.
	 * @param doSort true if the items should be sorted.
	 */
	public void prepareMenuBlock(String menuBlockName, boolean doSort) {
		if (contentsMap.containsKey(menuBlockName)) {
			return;
		}
		MenuBlockContent menuBlockContent = new MenuBlockContent(menuBlockName, doSort);
		contentsMap.put(menuBlockName, menuBlockContent);
		contentsList.add(menuBlockContent);
	}
	
	/**
	 * Adds the menu item to the given menu block. If no menu block of this name has been prepared at this point then
	 * the preparation is done first.
	 * 
	 * @param menuBlockName The name of the menu block.
	 * @param menuItem The menu item to be added to the menu block.
	 */
	public void addMenuItem(String menuBlockName, MenuItem menuItem) {
		if (menuItem instanceof MenuEntry) {
			MenuEntry me = (MenuEntry) menuItem;
			me.getTextElement().setID(id);
			boolean restrFulfilled = me.getTextElement().applyRestriction(restriction);
			if (!restrFulfilled) return;
		}
		if (!contentsMap.containsKey(menuBlockName)) {
			prepareMenuBlock(menuBlockName);
		}
		contentsMap.get(menuBlockName).addItem(menuItem);
	}
	
	/**
	 * Adds a new menu entry containing the text element to the menu block. If no menu block of this name has been
	 * prepared at this point then the preparation is done first.
	 * 
	 * @param menuBlockName The name of the menu block.
	 * @param textElement The text element of the menu entry that is created and then added to the menu block.
	 */
	public void addMenuEntry(String menuBlockName, TextElement textElement) {
		textElement.setID(id);
		boolean restrFulfilled = textElement.applyRestriction(restriction);
		if (!restrFulfilled) return;
		if (!contentsMap.containsKey(menuBlockName)) {
			prepareMenuBlock(menuBlockName);
		}
		contentsMap.get(menuBlockName).addEntry(textElement);
	}
	
	synchronized List<MenuBlockContent> createMenu(ChartParser parser) {
		contentsMap.clear();
		contentsList.clear();
		
		initMenuCreation();
		
		for (Restriction r : parser.getNextTokenRestrictions()) {
			restriction = r;
			id = parser.getTokens().size();
			addMenuItems(r);
		}
		
		return new ArrayList<MenuBlockContent>(contentsList);
	}
	
	/**
	 * Adds menu entries that stand for variables in the form of "X", "Y", "Z", "X1", "Y1", and so on.
	 * 
	 * @param menuBlockName The name of the menu block into which the entries should be added.
	 * @param num The number upto which the variables should be enumerated.
	 * @param categoryName The category name of the entries.
	 */
	public void addVariableEntries(String menuBlockName, int num, String categoryName) {
		ArrayList<String> varNames = new ArrayList<String>();
		varNames.add("X");
		varNames.add("Y");
		varNames.add("Z");
		for (int i=1; i<=num; i++) {
			varNames.add("X" + i);
			varNames.add("Y" + i);
			varNames.add("Z" + i);
		}
		for (String s : varNames) {
			Terminal t = new Terminal(categoryName);
			t.setFeature("text", s);
			addMenuEntry(menuBlockName, new TextElement(s, t));
		}
	}

}
