// This file is part of the Attempto Java Packages.
// Copyright 2008-2009, Attempto Group, University of Zurich (see http://attempto.ifi.uzh.ch).
//
// The Attempto Java Packages is free software: you can redistribute it and/or modify it under the
// terms of the GNU Lesser General Public License as published by the Free Software Foundation,
// either version 3 of the License, or (at your option) any later version.
//
// The Attempto Java Packages is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE. See the GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License along with the Attempto
// Java Packages. If not, see http://www.gnu.org/licenses/.

package ch.uzh.ifi.attempto.preditor.text;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import ch.uzh.ifi.attempto.chartparser.Restriction;
import ch.uzh.ifi.attempto.chartparser.Terminal;

/**
 * This class describes a text element (a word or phrase) that is used by the
 * predictive editor. Every text element contains a text and one or more grammatical categories.
 * If a text element contains more than one categories then this means that it can stand
 * for any of these categories.
 * 
 * @author Tobias Kuhn
 */
public class TextElement {
	
	private String text;
	private ArrayList<Terminal> categories = new ArrayList<Terminal>();
	private TextContainer textContainer;
	
	/**
	 * Creates a new text element with one category.
	 * 
	 * @param text The text.
	 * @param category The category.
	 */
	public TextElement(String text, Terminal category) {
		this.text = text;
		this.categories.add(category);
	}
	
	/**
	 * Creates a new text element with the given categories.
	 * 
	 * @param text The text.
	 * @param categories The categories.
	 */
	public TextElement(String text, List<Terminal> categories) {
		this.text = text;
		this.categories.addAll(categories);
	}
	
	/**
	 * Creates a new text element.
	 * 
	 * @param text The text.
	 * @param categoryName The name of the category.
	 */
	public TextElement(String text, String categoryName) {
		this(text, new Terminal(categoryName));
	}
	
	/**
	 * Creates a new text element where the text is the name of the category.
	 * 
	 * @param category The category.
	 */
	public TextElement(Terminal category) {
		this(category.getName(), category);
	}
	
	/**
	 * Creates a new text element where the text is the name of the category.
	 * 
	 * @param categoryName The name of the category.
	 */
	public TextElement(String categoryName) {
		this(categoryName, new Terminal(categoryName));
	}
	
	/**
	 * Creates a text element with an empty string as its text and category of the same name.
	 */
	public TextElement() {
		this("");
	}
	
	/**
	 * Creates a new text element containing a number which is also the name of the category.
	 * 
	 * @param number The number.
	 */
	public TextElement(int number) {
		this(number + "", new Terminal(number + ""));
	}
	
	/**
	 * Returns the original text of this text element. This is the unchanged text before it is given
	 * to the context checker.
	 * 
	 * @return The original text (before the context checker).
	 */
	public String getOriginalText() {
		return text;
	}
	
	/**
	 * Returns the text of this text element. The text might have been changed by the context checker.
	 * 
	 * @return The text (after the context checker).
	 */
	public String getText() {
		if (textContainer != null) {
			return textContainer.getTextElementText(this);
		}
		return getOriginalText();
	}
	
	/**
	 * Returns the grammatical category of this text element.
	 * 
	 * @return The grammatical category.
	 */
	public Terminal[] getCategories() {
		return (Terminal[]) categories.toArray(new Terminal[0]);
	}
	
	void setTextContainer(TextContainer textContainer) {
		this.textContainer = textContainer;
	}
	
	void removeTextContainer() {
		this.textContainer = null;
	}
	
	/**
	 * Applies the given restriction by excluding all categories that are not fulfilled by the
	 * restriction. The return value indicates whether this text elements fulfills the
	 * restriction which is the case if and only if at least one category is left. 
	 * 
	 * @param restr The restriction to be applied.
	 * @return true if the restrictions is altogether fulfilled, i.e. if at least one category
	 *   fulfills the restriction.
	 */
	public final boolean applyRestriction(Restriction restr) {
		List<Terminal> categories = new ArrayList<Terminal>(Arrays.asList(getCategories()));
		if (restr == null) return true;
		for (Terminal t : new ArrayList<Terminal>(categories)) {
			if (!restr.isFulfilledBy(t)) categories.remove(t);
		}
		return (!categories.isEmpty());
	}
	
	/**
	 * Sets the feature "id" for all categories.
	 * 
	 * @param id The id value.
	 */
	public void setID(int id) {
		for (Terminal t : categories) {
			t.setFeature("id", id + "");
		}
	}
	
	/**
	 * Includes the categories of the given text element. This is only allowed if the two
	 * text elements are equal (accoring to the equals-method).
	 * 
	 * @param textElement The text element whose categories should be included.
	 */
	public void include(TextElement textElement) {
		if (!equals(textElement)) {
			throw new RuntimeException("Only equal text elements can be included");
		}
		categories.addAll(textElement.categories);
	}
	
	/**
	 * Two text elements are equals if they share the same text.
	 */
	public boolean equals(Object obj) {
		if (obj instanceof TextElement) {
			TextElement other = (TextElement) obj;
			return (this.text.equals(other.text));
		}
		return false;
	}
	
	public String toString() {
		return text;
	}

}
